@@ -5,7 +5,7 @@ Configuration handling for OpenEvolve
 import os
 from dataclasses import dataclass, field
 from pathlib import Path
-from typing import Any, Dict, List, Optional, Union
+from typing import Any, Callable, Dict, List, Optional, Union
 
 import yaml
 
@@ -19,6 +19,9 @@ class LLMModelConfig:
     api_key: Optional[str] = None
     name: str = None
 
+    # Custom LLM client
+    init_client: Optional[Callable] = None
+
     # Weight for model in ensemble
     weight: float = 1.0
 
@@ -35,6 +38,9 @@ class LLMModelConfig:
 
     # Reproducibility
     random_seed: Optional[int] = None
+    
+    # Reasoning parameters
+    reasoning_effort: Optional[str] = None
 
 
 @dataclass
@@ -66,6 +72,9 @@ class LLMConfig(LLMModelConfig):
     primary_model_weight: float = None
     secondary_model: str = None
     secondary_model_weight: float = None
+    
+    # Reasoning parameters (inherited from LLMModelConfig but can be overridden)
+    reasoning_effort: Optional[str] = None
 
     def __post_init__(self):
         """Post-initialization to set up model configurations"""
@@ -118,6 +127,7 @@ class LLMConfig(LLMModelConfig):
             "retries": self.retries,
             "retry_delay": self.retry_delay,
             "random_seed": self.random_seed,
+            "reasoning_effort": self.reasoning_effort,
         }
         self.update_model_params(shared_config)
 
@@ -170,6 +180,7 @@ class LLMConfig(LLMModelConfig):
             "retries": self.retries,
             "retry_delay": self.retry_delay,
             "random_seed": self.random_seed,
+            "reasoning_effort": self.reasoning_effort,
         }
         self.update_model_params(shared_config)
 
@@ -302,6 +313,19 @@ class EvaluatorConfig:
     max_artifact_storage: int = 100 * 1024 * 1024  # 100MB per program
 
 
+@dataclass
+class EvolutionTraceConfig:
+    """Configuration for evolution trace logging"""
+    
+    enabled: bool = False
+    format: str = "jsonl"  # Options: "jsonl", "json", "hdf5"
+    include_code: bool = False
+    include_prompts: bool = True
+    output_path: Optional[str] = None
+    buffer_size: int = 10
+    compress: bool = False
+
+
 @dataclass
 class Config:
     """Master configuration for OpenEvolve"""
@@ -319,6 +343,7 @@ class Config:
     prompt: PromptConfig = field(default_factory=PromptConfig)
     database: DatabaseConfig = field(default_factory=DatabaseConfig)
     evaluator: EvaluatorConfig = field(default_factory=EvaluatorConfig)
+    evolution_trace: EvolutionTraceConfig = field(default_factory=EvolutionTraceConfig)
 
     # Evolution settings
     diff_based_evolution: bool = True
@@ -344,7 +369,7 @@ class Config:
 
         # Update top-level fields
         for key, value in config_dict.items():
-            if key not in ["llm", "prompt", "database", "evaluator"] and hasattr(config, key):
+            if key not in ["llm", "prompt", "database", "evaluator", "evolution_trace"] and hasattr(config, key):
                 setattr(config, key, value)
 
         # Update nested configs
@@ -367,6 +392,8 @@ class Config:
             config.database.random_seed = config.random_seed
         if "evaluator" in config_dict:
             config.evaluator = EvaluatorConfig(**config_dict["evaluator"])
+        if "evolution_trace" in config_dict:
+            config.evolution_trace = EvolutionTraceConfig(**config_dict["evolution_trace"])
 
         return config
 
@@ -435,6 +462,15 @@ class Config:
                 "use_llm_feedback": self.evaluator.use_llm_feedback,
                 "llm_feedback_weight": self.evaluator.llm_feedback_weight,
             },
+            "evolution_trace": {
+                "enabled": self.evolution_trace.enabled,
+                "format": self.evolution_trace.format,
+                "include_code": self.evolution_trace.include_code,
+                "include_prompts": self.evolution_trace.include_prompts,
+                "output_path": self.evolution_trace.output_path,
+                "buffer_size": self.evolution_trace.buffer_size,
+                "compress": self.evolution_trace.compress,
+            },
             # Evolution settings
             "diff_based_evolution": self.diff_based_evolution,
             "max_code_length": self.max_code_length,
